import sys
import os
import platform

pytest_plugins = ['helpers_namespace']

import pytest
from pathlib import Path
from subprocess import Popen, PIPE


if sys.version_info < (3, 4):
    print("Requires Python 3.4+")
    sys.exit(1)


TESTS_ROOT = os.path.abspath(os.path.dirname(__file__))
PROJECT_ROOT = os.path.dirname(TESTS_ROOT)


@pytest.fixture
def resources():
    return Path(TESTS_ROOT) / 'resources'


@pytest.fixture(scope="function")
def outdir(tmpdir):
    return Path(str(tmpdir))



import pytest
import imghdr
from io import BytesIO
from PIL import Image
import zlib
import sys

from pikepdf import Pdf, Object, PdfImage
from pikepdf._qpdf import Name, Null


@pytest.fixture
def congress(resources):
    pdf = Pdf.open(resources / 'congress.pdf')
    pdfimage = pdf.pages[0].Resources.XObject['/Im0']
    return pdfimage, pdf


def test_image_from_nonimage(resources):
    pdf = Pdf.open(resources / 'congress.pdf')
    resources = pdf.pages[0].Contents
    with pytest.raises(TypeError):
        PdfImage(resources)


def test_image(congress):
    pdfimage = PdfImage(congress[0])
    pillowimage = pdfimage.topil()

    assert pillowimage.mode == pdfimage.mode
    assert pillowimage.size == pdfimage.size


def test_image_replace(congress, outdir):
    pdfimage = PdfImage(congress[0])
    pillowimage = pdfimage.topil()

    grayscale = pillowimage.convert('L')

    congress[0].write(zlib.compress(grayscale.tobytes()), Name("/FlateDecode"), Null())
    congress[0].ColorSpace = Name("/DeviceGray")
    pdf = congress[1]
    pdf.save(outdir / 'congress_gray.pdf')


def test_lowlevel_jpeg(congress, outdir):
    raw_bytes = congress[0].read_raw_bytes()
    with pytest.raises(RuntimeError):
        congress[0].read_bytes()

    assert imghdr.what('', h=raw_bytes) == 'jpeg'

    im = Image.open(BytesIO(raw_bytes))
    assert im.size == (congress[0].Width, congress[0].Height)
    assert im.mode == 'RGB'


def test_lowlevel_replace_jpeg(congress, outdir):
    # This test will modify the PDF so needs its own image
    raw_bytes = congress[0].read_raw_bytes()

    im = Image.open(BytesIO(raw_bytes))
    grayscale = im.convert('L')

    congress[0].write(zlib.compress(grayscale.tobytes()), Name("/FlateDecode"), Null())
    congress[0].ColorSpace = Name('/DeviceGray')

    pdf = congress[1]
    pdf.save(outdir / 'congress_gray.pdf')


# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (C) 2017, James R. Barlow (https://github.com/jbarlow83/)


from collections import namedtuple
from pkg_resources import get_distribution, DistributionNotFound

import os

try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    __version__ = "Not installed"
    pass
try:
    from . import _qpdf
except ImportError:
    raise ImportError("pikepdf's extension library failed to import")

from ._qpdf import Object, ObjectType, PdfError, Pdf, PasswordError, \
    ObjectStreamMode, StreamDataMode, Boolean, Integer, Real, Name, String, \
    Array, Dictionary, Stream, Operator, Null

from ._pdfimage import PdfImage

__libqpdf_version__ = _qpdf.qpdf_version()


class _OperandGrouper(_qpdf.StreamParser):
    """Parse a PDF content stream into a sequence of instructions.

    Helper class for parsing PDF content streams into instructions. Semantics
    are a little weird since it is subclassed from C++.

    """

    PdfInstruction = namedtuple('PdfInstruction', ('operands', 'operator'))

    def __init__(self):
        super().__init__()
        self.instructions = []
        self._tokens = []

    def handle_object(self, obj):
        if obj.type_code == ObjectType.operator:
            instruction = self.PdfInstruction(
                operands=self._tokens, operator=obj)
            self.instructions.append(instruction)
            self._tokens = []
        else:
            self._tokens.append(obj)

    def handle_eof(self):
        if self._tokens:
            raise EOFError("Unexpected end of stream")


def parse_content_stream(stream):
    """Parse a PDF content stream into a sequence of instructions.

    A PDF content stream is list of instructions that describe where to render
    the text and graphics in a PDF. This is the starting point for analyzing
    PDFs.

    Each instruction contains at least one operator and zero or more operands.

    The `stream` object may be either a `Object.Stream` type or an array of
    streams.

    >>> pdf = pikepdf.Pdf.open(input_pdf)
    >>> stream = pdf.pages[0].Contents
    >>> for operands, command in parse_content_stream(stream):
    >>>     print(command)

    """

    if not isinstance(stream, Object):
        raise TypeError("stream must a PDF object")

    grouper = _OperandGrouper()
    try:
        Object.parse_stream(stream, grouper)
    except PdfError as e:
        if 'parseContentStream called on non-stream' in str(e):  # qpdf 6.x
            raise TypeError("parse_content_stream called on non-stream Object")
        elif 'ignoring non-stream while parsing' in str(e):  # qpdf 7.0
            raise TypeError("parse_content_stream called on non-stream Object")
        raise e from e

    return grouper.instructions


class Page:
    def __init__(self, obj):
        self.obj = obj

    def __getattr__(self, item):
        return getattr(self.obj, item)

    def __setattr__(self, item, value):
        if item == 'obj':
            object.__setattr__(self, item, value)
        elif hasattr(self.obj, item):
            setattr(self.obj, item, value)
        else:
            raise AttributeError(item)

    def __repr__(self):
        return repr(self.obj).replace(
            'pikepdf.Dictionary', 'pikepdf.Page', 1)

    @property
    def mediabox(self):
        return self.obj.MediaBox

    def has_text(self):
        """Check if this page print text

        Search the content stream for any of the four text showing operators.
        We ignore text positioning operators because some editors might
        generate maintain these even if text is deleted etc.

        This cannot detect raster text (text in a bitmap), text rendered as
        curves. It also cannot determine if the text is visible to the user.

        :return: True if there is text
        """
        text_showing_operators = {Operator(op) for op in
                                  ('Tj', 'TJ', '"', "'")}
        for _, operator in parse_content_stream(self.obj.Contents):
            if operator in text_showing_operators:
                return True
        return False





import gc
import pytest

from pikepdf import Pdf

# Try to do some things without blowing up

def test_access_image(resources):
    pdf = Pdf.open(resources / 'congress.pdf')
    im0 = pdf.pages[0].Resources.XObject['/Im0']
    del pdf
    gc.collect()
    im0.read_raw_bytes()


def test_access_page(resources):
    pdf = Pdf.open(resources / 'graph.pdf')
    page0 = pdf.pages[0]
    del pdf
    gc.collect()
    page0.Contents.read_raw_bytes()


def test_remove_pdf_and_all_pages(resources):
    pdf = Pdf.open(resources / 'graph.pdf')
    page0 = pdf.pages[0]
    contents = page0.Contents
    del pdf
    del page0
    gc.collect()
    contents.read_raw_bytes()


def test_access_pdf_metadata(resources):
    pdf = Pdf.open(resources / 'graph.pdf')
    meta = pdf.Root.Metadata
    del pdf
    gc.collect()
    meta.read_raw_bytes()